import asyncio
import oneai

from typing import Awaitable, Dict, Iterable, List, Union
from oneai.classes import Input, Output, Skill


class Pipeline:
    def __init__(self, steps: List[Skill], api_key: str=None) -> None:
        self.steps = steps  # todo: validate (based on input_type)
        self.api_key = api_key

    def to_json(self) -> dict:
        result = []
        for skill in self.steps:
            result.append({
                'skill': skill.api_name,
                'params': {
                    p: skill.__getattribute__(p) for p in skill._skill_params
                }
            })
        return result
        
    def run(
        self,
        input: Union[str, Input, Iterable[Union[str, Input]]],
        api_key: str=None
    ) -> Output:
        return asyncio.run(self.run_async(input, api_key))

    async def run_async(
        self,
        input: Union[str, Input, Iterable[Union[str, Input]]],
        api_key: str=None
    ) -> Awaitable[Output]:
        if isinstance(input, (str, Input)):
            from oneai.requests import send_single_request
            return await send_single_request(
                input,
                self,
                api_key=api_key or self.api_key or oneai.api_key,
            )
        elif isinstance(input, Iterable):
            return await self.run_batch_async(input, api_key)
        else: raise TypeError(f'pipeline input must be Input, str or iterable of inputs')

    def run_batch(
        self,
        batch: Iterable[Union[str, Input]],
        api_key: str=None
    ) -> Dict[Union[str, Input], Output]:
        return asyncio.run(self.run_batch_async(batch, api_key))

    async def run_batch_async(
        self,
        batch: Iterable[Union[str, Input]],
        api_key: str=None
    ) -> Awaitable[Dict[Union[str, Input], Output]]:
        from oneai.requests import send_batch_request
        return await send_batch_request(
            batch,
            self,
            api_key=api_key or self.api_key or oneai.api_key
        )

    def __repr__(self) -> str:
        return f'oneai.Pipeline({self.steps})'



from copy import copy
from dataclasses import dataclass, field
import json
from typing import Iterable, List, Tuple, Type, Union

_SkillInput = Type[Union[str, 'Input']]
_SkillOutput = Type[Tuple[Union[str, 'Input'], str]]


@dataclass
class Skill:
    '''
    Args:
        api_name: The name of the Skill in the pipeline API.
        is_generator: Whether the Skill is a generator.
        skill_params: Names of the fields of the Skill object that should be passed as parameters to the API.
        label_type: If the Skill generates labels, the type name of the label.
        output_attr: The attribute name of the Skill's output in the Output object.
        output_attr1: Only for Skills with 2 outputs (text / labels)
    '''
    api_name: str = ''
    is_generator: bool = False
    _skill_params: List[str] = field(default_factory=list, repr=False, init=False)
    # todo: replace all these w/ an output type object + parse conversations from t. enhancer etc.
    label_type: str = ''
    output_attr: str = ''
    output_attr1: str = ''
    

def skillclass(
    cls: Type=None,
    api_name: str='',
    label_type: str='',
    is_generator: bool=False,
    output_attr: str = '',
    output_attr1: str = ''
):
    def wrap(cls) -> cls:
        if not issubclass(cls, Skill):
            print(f'warning: class {cls.__name__} decorated with @skillclass does not inherit Skill')

        def __init__(self, *args, **kwargs):
            cls_init(self, *args, **kwargs)
            Skill.__init__(self, api_name=api_name, label_type=label_type, is_generator=is_generator, output_attr=output_attr, output_attr1=output_attr1)
            self._skill_params = [a for a in self.__dict__ if not (a in Skill.__dict__ or a == '_skill_params')]
        
        cls_init = cls.__init__
        cls.__init__ = __init__
        return cls
    
    return wrap if cls is None else wrap(cls)


class Input:
    def __init__(self, type: str):
        self.type = type

    @classmethod
    def parse(cls, text: str) -> 'Input':
        raise NotImplementedError()

    def get_text(self) -> str:
        raise NotImplementedError()

class Document(Input):
    def __init__(self, text: str):
        super().__init__('article')
        self.text = text

    @classmethod
    def parse(cls, text: str) -> 'Document':
        return cls(text)

    def get_text(self) -> str:
        return self.text

@dataclass
class Utterance:
    speaker: str
    utterance: str

    def __repr__(self) -> str:
        return f'\n\t{self.speaker}: {self.utterance}'

class Conversation(Input):
    def __init__(self, utterances: List[Utterance]=[]):
        super().__init__('conversation')
        self.utterances = utterances

    def get_text(self) -> str:
        return json.dumps(self.utterances, default=lambda o: o.__dict__)

    @classmethod
    def parse(cls, text: str) -> 'Conversation':
        try:
            js = json.loads(text)
            return cls([Utterance(**utterance) for utterance in js])
        except json.JSONDecodeError:
            from oneai.parsing import parse_conversation
            return parse_conversation(text)

    def __repr__(self) -> str:
        return f'oneai.Conversation{repr(self.utterances)}'

@dataclass
class Label:
    type: str = ''
    name: str = ''
    span: List[int] = field(default_factory=lambda: [0, 0])
    value: str = ''
    data: dict = field(default_factory=lambda: dict())

    @classmethod
    def from_json(cls, object): return cls(
        type=object.get('type', ''),
        name=object.get('name', ''),
        span=object.get('span', [0, 0]),
        value=object.get('value', ''),
        data=object.get('data', {})
    )

    def __repr__(self) -> str:
        return 'oneai.Label(' + ', '.join(f'{k}={v}' for k, v in self.__dict__.items() if v) + ')'

@dataclass
class Output:
    text: str
    skills: List[Skill] # not a dict since Skills are currently mutable & thus unhashable
    data: List[Union[List[Label], 'Output']]

    def __getitem__(self, name: str) -> Union[List[Label], 'Output']:
        return self.__getattr__(name)

    def __getattr__(self, name: str) -> Union[List[Label], 'Output']:
        #####TEMP#HACK########
        if name == 'business_entities' and hasattr(self, 'labs'):
            return self.__getattr__('labs').business_entities
        ######################
        for i, skill in enumerate(self.skills):
            if (skill.api_name and skill.api_name == name) or \
                (skill.output_attr and name in skill.output_attr) or \
                (type(skill).__name__ == name):
                return self.data[i]
        raise AttributeError(f'{name} not found in {self}')

    def __dir__(self) -> Iterable[str]:
        return super().__dir__() + [skill.output_attr or skill.api_name for skill in self.skills]

    @classmethod
    def build(
        cls,
        pipeline, 
        raw_output: dict,
        input_type: type=str
    ) -> 'Output':
        def get_text(index, input_type):
            # get the input text for this Output object. use index=-1 to get the original input text
            # text can be returned as a simple str or parsed to match a given input type
            text = raw_output['output'][index]['text'] if index >= 0 else raw_output['input_text']
            return input_type.parse(text) if issubclass(input_type, Input) else text

        def split_pipeline(skills, i):
            # split pipeline at a generator Skill
            first, second = skills[:i + 1], skills[i + 1:]
            if hasattr(skills[i], 'output_attr1') and skills[i].output_attr1: 
                # handle skills that create both text and labels
                clone = copy(skills[i])
                clone.is_generator = False
                clone.output_attr = skills[i].output_attr1
                second.insert(0, clone)
            return first, second

        def _build_internal(
            output_index: int, 
            skills: List[Skill],
            input_type: type
        ) -> 'Output':
            text = get_text(output_index, input_type)
            labels = [Label.from_json(label) for label in raw_output['output'][output_index]['labels']]
            data = []
            for i, skill in enumerate(skills):
                if skill.is_generator:
                    skills, next_skills = split_pipeline(skills, i)
                    data.append(_build_internal(output_index + 1, next_skills, str))
                    break
                else:
                    data.append(list(filter(lambda label: label.type == skill.label_type, labels)))
            return cls(text=text, skills=skills, data=data)

        
        generator = raw_output['output'][0].get('text_generated_by_step_id', 0) - 1
        if generator < 0:
            return _build_internal(0, pipeline.steps, input_type)
        else:
            # edge case- first Skill is a generator, or a generator preceded by Skills that didn't generate output
            # in this case the API will skip these Skills,
            # so we need to create filler objects to match the expected structure
            skills, next_skills = split_pipeline(pipeline.steps, generator)
            return cls(
                text=get_text(-1, input_type),
                skills=skills,
                data=[[]] * generator + [_build_internal(0, next_skills, str)]
            )

    def __repr__(self) -> str:
        result = f'oneai.Output(text={repr(self.text)}'
        for i, skill in enumerate(self.skills):
            result += f', {skill.output_attr or skill.api_name}={repr(self.data[i])}'
        return result + ')'



from aiohttp import ClientResponse

# todo: input type validation errors


class InputError(Exception):
    pass

class APIKeyError(Exception):
    pass

class ServerError(Exception):
    pass

# todo: fix this
errors = {
    400: InputError,
    401: APIKeyError,
    403: APIKeyError,
    500: ServerError,
    503: ServerError
}

def handle_unsuccessful_response(response: ClientResponse):
    raise errors.get(response.status, ServerError)(response.reason)



import asyncio
from datetime import datetime, timedelta
from typing import Awaitable, Dict, Iterable, Union

import aiohttp
import oneai

from oneai.classes import Input, Output
from oneai.exceptions import handle_unsuccessful_response

MAX_CONCURRENT_REQUESTS = 4


async def _send_request(
    session: aiohttp.ClientSession,
    input: Union[Input, str],
    pipeline: oneai.Pipeline,
    api_key: str
) -> Awaitable[Output]:
    headers = {
        'api-key': api_key,
        'Content-Type': 'application/json'
    }
    request = {
        'text': input.get_text() if isinstance(input, Input) else str(input),
        'steps': pipeline.to_json(),
        'input_type': input.type if isinstance(input, Input) else 'article'
    }
    async with session.post(oneai.URL, headers=headers, json=request) as response:
        if response.status != 200: 
            handle_unsuccessful_response(response)
        else:
            return Output.build(pipeline, await response.json(), input_type=type(input))


async def send_single_request(
    input: Union[Input, str],
    pipeline: oneai.Pipeline,
    api_key: str
) -> Awaitable[Output]:
    timeout = aiohttp.ClientTimeout(total=6000)
    async with aiohttp.ClientSession(timeout=timeout) as session:
        return await _send_request(
            session,
            input,
            pipeline,
            api_key
        )


async def send_batch_request(
    batch: Iterable[Union[str, Input]],
    pipeline: oneai.Pipeline,
    api_key: str
) -> Awaitable[Dict[Union[str, Input], Output]]:
    iterator = iter(batch)
    results = dict()
    exceptions = 0
    time_total = timedelta()

    def next_input():
        try: return next(iterator)
        except StopIteration: return None  # we need to break loop for each worker, so we ignore StopIteration

    def print_progress(time_delta=timedelta(), start=False, end=False): # todo progress bar for iterables with __len__
        nonlocal time_total
        
        def time_format(time: timedelta):
            minutes = f'{time.seconds // 60}m ' if time.seconds > 59 else ''
            return minutes + f'{time.seconds % 60}s {time.microseconds // 1000}ms'

        total = len(results)
        time_total += time_delta
        if start: print(f'Starting processing batch with {MAX_CONCURRENT_REQUESTS} workers', end='\r')
        elif end: print('Processed %d inputs - %s/input - %s total - %d successful - %d failed' % (
            total,
            time_format(time_delta),
            time_format(time_total),
            total - exceptions,
            exceptions
        ))
        else: print('Input %d - %s/input - %s total - %d successful - %d failed        ' % (
            total,
            time_format(time_delta),
            time_format(time_total),
            total - exceptions,
            exceptions
        ), end='\r')

    async def req_worker(session):
        nonlocal exceptions

        time_start = datetime.now()
        input = next_input()
        while input:
            try: results[input] = await _send_request(
                session,
                input,
                pipeline,
                api_key
            )
            except Exception as e:
                print(f'\r\033[KInput {len(results)}:', repr(e))
                results[input] = e
                exceptions += 1

            time_end = datetime.now()
            if oneai.PRINT_PROGRESS:
                print_progress(time_end - time_start)
            time_start = time_end
            input = next_input()

    timeout = aiohttp.ClientTimeout(total=6000)
    workers = []
    async with aiohttp.ClientSession(timeout=timeout) as session:
        for _ in range(MAX_CONCURRENT_REQUESTS):
            worker = asyncio.create_task(req_worker(session))
            workers.append(worker)
        print_progress(start=True)
        await asyncio.gather(*workers)
        print_progress(end=True)
        return results



import re
import oneai


def parse_conversation(text: str, strict=False) -> oneai.Conversation:
    result = []
    lines = re.split(r'\r?\n', text.strip())
    firstLine = True
    structure = None
    currentLineInfo = None
    waitForTextLine = False
    weak = False
    
    for i, line in enumerate(lines):
        if _isEmptyOrWhitespace(line):
            continue

        if waitForTextLine:
            result.append(oneai.Utterance(
                speaker=currentLineInfo['speaker'],
                utterance=line.strip()
            ))
            waitForTextLine = False
            continue
            
        currentLineInfo = _parseSpeakerLine(line)
        if currentLineInfo is None:
            raise ValueError(f'Invalid conversation format at line {i}')
        
        if firstLine:
            structure = currentLineInfo
            firstLine = False

        weak |= currentLineInfo['weak']
        if strict and not _comp(structure, currentLineInfo):
            raise ValueError(f'Differing conversation format at line {i}, run with strict=False to ignore')
        
        if currentLineInfo['text']:
            result.append(oneai.Utterance(
                speaker=currentLineInfo['speaker'],
                utterance=currentLineInfo['text']
            ))
            waitForTextLine = False
        else:
            waitForTextLine = True
    if waitForTextLine:
        raise ValueError('Invalid conversation format: last utterance is empty, at line')
    
    if weak and len(result) <= 3:
        raise ValueError('Weak pattern detected, but conversation is too short')

    return oneai.Conversation(utterances=result)

def _isEmptyOrWhitespace(text):
    return (not text) or (text.isspace())


def _parseSpeakerLine(text):
    value = {
        'weak': True,
        'preTime': False,
        'speaker': None,
        'time': False,
        'separator': False,
        'text': None
    }

    match = re.match(r'\s*\[?[0-9]+:[0-9]+\]?\s*', text)
    if match:
        text = text[match.end():]
        value['preTime'] = True
        value['weak'] = False

    # capture STRONG pattern - either timestamp or separator
    matchArea = text[:min(len(text), 35)]
    match = re.search(r'\s{1,10}\[?[0-9]{1,2}:[0-9]{1,2}\]?\s*|\s{0,5}[:|-]', matchArea)

    if not match: # STRONG not found
        # check if speaker only, in all caps - WEAK PATTERN
        match = re.match(r'[A-Z_-]{3,20}$', matchArea)
        if not match: return None # not a valid format - fail
        else: value['speaker'] = match.group(0)
    else:
        value['weak'] = False
        if match.end() < len(text):
            value['text'] = text[match.end():].strip()
        value['speaker'] = text[:match.start()]

        if re.search(r'[:|-]$', match.group(0)):
            value['separator'] = True
        else:
            text = text[match.end():]
            match = re.match(r'\s*[:|-]', text)
            if match:
                value['separator'] = True
                value['text'] = text[match.end():].strip() if match.end() < len(text) else None
            value['time'] = True
    return value


def _comp(a, b):
    return a['separator'] == b['separator'] and \
            a['time'] == b['time'] and \
            a['preTime'] == b['preTime'] and \
            (a['text'] is None) == (b['text'] is None)



from dataclasses import dataclass

from oneai.classes import Skill, skillclass

@skillclass(api_name='enhance', is_generator=True, label_type='replacement', output_attr='enhanced', output_attr1='replacements')
class TranscriptionEnhancer(Skill): pass

@skillclass(api_name='summarize', is_generator=True, label_type='origin', output_attr='summary', output_attr1='origins')
@dataclass
class Summarize(Skill):
    '''
    Provides a summary of the input

    Attributes
    ----------
        min_length : int, default=5
            minimal desired length (words) of the summary 
        max_length : int, default=100
            maximal desired length (words) of the summary 


    Examples
    --------
    >>> pipeline = oneai.Pipeline(steps=[
    ...     oneai.skills.HTMLExtractArticle(),
    ...     oneai.skills.Summarize(min_length=10)
    ... ])
    >>> pipeline.run('https://www.oneai.com/docs/')
    '''
    min_length: int = 5
    max_length: int = 100
    find_origins: bool = True

@skillclass(api_name='emotions', label_type='emotion')
class Emotions(Skill): pass

@skillclass(api_name='entities', label_type='entity')
class Entities(Skill): pass

@skillclass(api_name='keywords', label_type='keyword')
class Keywords(Skill): pass

@skillclass(api_name='highlights', label_type='highlight')
class Highlights(Skill): pass

@skillclass(api_name='sentiments', label_type='sentiment')
class Sentiments(Skill): pass

@skillclass(api_name='article-topics', label_type='topic', output_attr='topics')
class Topics(Skill): pass

@skillclass(api_name='extract-html')
class HTMLExtractArticle(Skill): pass

@skillclass(api_name='html-extract-text')
class HTMLExtractText(Skill): pass

@skillclass(api_name='business-entities', is_generator=True, label_type='business-entity', output_attr='labs', output_attr1='business_entities')
class BusinessEntities(Skill): pass

'''[ { "text_generated_by_step_name": "business-entities", "text_generated_by_step_id": 1, "text": 
"labels": [ { "type": "business-entity", "name": "quantity", "span": [ 1022, 1054 ], "value": "25,000 Barrels plus or minus 10%", "output_spans": [ { "section": 0, "start": 1022, "end": 1054 } ], "input_spans": null, "span_text": "25,000 Barrels plus or minus 10%", "data": { "quantity": "25,000", "modifier": "plus or minus 10 %", "unit": "Barrels", "secondary_unit": "", "is_primary_confidence": 1 } }, { "type": "business-entity", "name": "pricing", "span": [ 1473, 1503 ], "value": "Differential: USD 0.09 per GAL", "output_spans": [ { "section": 0, "start": 1473, "end": 1503 } ], "input_spans": null, "span_text": "Differential: USD 0.09 per GAL", "data": { "amount": "0.09", "currency": "USD", "modifier": "Differential :", "unit": "GAL", "is_primary_confidence": 1 } }, { "type": "business-entity", "name": "pricing", "span": [ 1778, 1808 ], "value": "Differential: USD 0.09 per GAL", "output_spans": [ { "section": 0, "start": 1778, "end": 1808 } ], "input_spans": null, "span_text": "Differential: USD 0.09 per GAL", "data": { "amount": "0.09", "currency": "USD", "modifier": "Differential :", "unit": "GAL", "is_primary_confidence": 0.33333333333333337 } }, { "type": "business-entity", "name": "pricing", "span": [ 1921, 1945 ], "value": "USD USD 0.007 per gallon", "output_spans": [ { "section": 0, "start": 1921, "end": 1945 } ], "input_spans": null, "span_text": "USD USD 0.007 per gallon", "data": { "amount": "0.007", "currency": "USD USD", "modifier": "", "unit": "gallon", "is_primary_confidence": 0.33333333333333337 } } ] } ]'''